/* PIC18F13K50 - Remote Messadapter
 * Nimmt verschiedene Messsignale auf und sendet diese dann in einem Paket ueber den UART
 * Empfngt ueber den UART Anweisungen wie z.B. Sample Rate etc.
 * 
 * Messeingnge:
 * 2x Strommessung I1/I2: AN7/AN9
 * 2x Sensormessung S1/S2: AN4/AN5
 * 2x Spannungsmessung V1/V2: AN6/AN10
 * 4x Logikmessung L1/L2/L3/L4: RB6/RC5/RC4/RC6
 * RC5 knnte auch fr Zeitmessungen ber T0Ckl (Timer0) oder CCP1 (Timer3) verwendet werden 
 * RC6 knnte auch fr Zeitmessungen ber T1OSCI (Timer1) oder T13CKI (Timer3) verwendet werden
 * 
 * UART-Konfig:
 * Positive Logik (Startbit=High=5V), 8 Datenbits, 1 Stopbit, no Parity, 38400 baud
 * 
 * Timer 2 (fr Sample-Rate):
 * Takt=3MHz (Fosc/4), Pre-Scale=1:16, Post-Scale=1:15, PR=49 -> d.h. berlauf alle 4ms/250Hz
 * Endgltige SampleRate erfolgt dann ber Variablen-Counter WaitCnt
 * 
 * Datenpaket an UART besteht aus 10 Byte (0..9):
 * Byte 0: Frame-Start = 0xFF
 * Byte 1: I1(7..0)
 * Byte 2: I2(5..0), I1(9..8)
 * Byte 3: S1(3..0), I2(9..6)
 * Byte 4: S2(1..0), S1(9..4)
 * Byte 5: S2(9..2)
 * Byte 6: V1(7..0)
 * Byte 7: V2(5..0), V1(9..8)
 * Byte 8: L4, L3, L2, L1, V2(9..6)
 * Byte 9: Frame-Stop = 0x00
 * 
 * Prozessor Takt = 3MHz : 12MHz Externer Quarz ohne PLL /4
 */

//Configuration-Bits - erstellt ber "Set Configuration Bits" - "Generate Source Code"
// CONFIG1L
#pragma config CPUDIV = NOCLKDIV// CPU System Clock Selection bits (No CPU System Clock divide)
#pragma config USBDIV = OFF     // USB Clock Selection bit (USB clock comes directly from the OSC1/OSC2 oscillator block; no divide)

// CONFIG1H
#pragma config FOSC = XT        // Oscillator Selection bits (XT oscillator)
#pragma config PLLEN = OFF      // 4 X PLL Enable bit (PLL is under software control)
#pragma config PCLKEN = ON      // Primary Clock Enable bit (Primary clock enabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       // Internal/External Oscillator Switchover bit (Oscillator Switchover mode disabled)

// CONFIG2L
#pragma config PWRTEN = ON      // Power-up Timer Enable bit (PWRT enabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable bits (Brown-out Reset disabled in hardware and software)
#pragma config BORV = 19        // Brown-out Reset Voltage bits (VBOR set to 1.9 V nominal)

// CONFIG2H
#pragma config WDTEN = OFF      // Watchdog Timer Enable bit (WDT is controlled by SWDTEN bit of the WDTCON register)
#pragma config WDTPS = 32768    // Watchdog Timer Postscale Select bits (1:32768)

// CONFIG3H
#pragma config HFOFST = OFF     // HFINTOSC Fast Start-up bit (The system clock is held off until the HFINTOSC is stable.)
#pragma config MCLRE = ON       // MCLR Pin Enable bit (MCLR pin enabled; RA3 input pin disabled)

// CONFIG4L
#pragma config STVREN = ON      // Stack Full/Underflow Reset Enable bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF        // Single-Supply ICSP Enable bit (Single-Supply ICSP disabled)
#pragma config BBSIZ = OFF      // Boot Block Size Select bit (512W boot block size)
#pragma config XINST = OFF      // Extended Instruction Set Enable bit (Instruction set extension and Indexed Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF        // Code Protection bit (Block 0 not code-protected)
#pragma config CP1 = OFF        // Code Protection bit (Block 1 not code-protected)

// CONFIG5H
#pragma config CPB = OFF        // Boot Block Code Protection bit (Boot block not code-protected)
#pragma config CPD = OFF        // Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       // Table Write Protection bit (Block 0 not write-protected)
#pragma config WRT1 = OFF       // Table Write Protection bit (Block 1 not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers not write-protected)
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot block not write-protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      // Table Read Protection bit (Block 0 not protected from table reads executed in other blocks)
#pragma config EBTR1 = OFF      // Table Read Protection bit (Block 1 not protected from table reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      // Boot Block Table Read Protection bit (Boot block not protected from table reads executed in other blocks)

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

#include <xc.h>


//Prototypen fr Funktionen
void GPIO_Init(void);
void ADC_Init(void);
void TMR2_Init(void);
void UART_Init(void);
void UART_Receive(void);
void UART_SendStopSeq(void);

unsigned char Flags;    //Verschiedene Flags (Bit0=Run/Stop)
unsigned char I1_L;     //Die Messdaten
unsigned char I1_H; 
unsigned char I2_L; 
unsigned char I2_H; 
unsigned char S1_L; 
unsigned char S1_H; 
unsigned char S2_L; 
unsigned char S2_H; 
unsigned char V1_L; 
unsigned char V1_H; 
unsigned char V2_L; 
unsigned char V2_H; 
unsigned char LB; 
unsigned char LC; 
unsigned char Byte;    //Das jeweilige bertragungs-Byte
unsigned char WaitCnt;  //Anzahl der TMR2-berlaufe die gewartet werden soll bis zur nchsten Samplung
unsigned char wait;
unsigned char b;        //Zhler


void main(void) { 
    Flags=0;        //Zum Start noch kein Run
    WaitCnt=10;     //Standard ist 10Hz sample rate
    GPIO_Init();
    ADC_Init();
    UART_Init();
    TMR2_Init();    //Setzt PR2 auf 50Hz und schaltet gleichzeitig TMR2 ein
    
    while(1){  
        wait=WaitCnt;
        if (Flags & 0x01) {     //Wenn Run-Flag gestzt ist
            //Messdaten einholen
            //I1 (AN7) messen
            ADCON0bits.CHS = 7;
            ADCON0bits.GO = 1;      //Ackquisition time wird durch gesetztes ACQT automatisch gewartet
            while (ADCON0bits.GO == 1);
            I1_L = ADRESL;
            I1_H = ADRESH;
            //I2 (AN9) messen
            ADCON0bits.CHS = 9;
            ADCON0bits.GO = 1;
            while (ADCON0bits.GO == 1);
            I2_L = ADRESL;
            I2_H = ADRESH;
            //S1 (AN4) messen
            ADCON0bits.CHS = 4;
            ADCON0bits.GO = 1;
            while (ADCON0bits.GO == 1);
            S1_L = ADRESL;
            S1_H = ADRESH;
            //S2 (AN5) messen
            ADCON0bits.CHS = 5;
            ADCON0bits.GO = 1;
            while (ADCON0bits.GO == 1);
            S2_L = ADRESL;
            S2_H = ADRESH;
            //V1 (AN6) messen
            ADCON0bits.CHS = 6;
            ADCON0bits.GO = 1;
            while (ADCON0bits.GO == 1);
            V1_L = ADRESL;
            V1_H = ADRESH;
            //V2 (AN10) messen
            ADCON0bits.CHS = 10;
            ADCON0bits.GO = 1;
            while (ADCON0bits.GO == 1);
            V2_L = ADRESL;
            V2_H = ADRESH;
            //RC-Status festhalten
            LC = PORTC;
            //RB-Status festhalten
            LB = PORTB;
            
            //Messdaten senden
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = 0xFF;           //Frame-Start = 0xFF senden
            Byte = I1_L;            //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 1: I1(7..0)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte1 senden
            Byte = (I2_L << 2) + I1_H;           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 2: I2(5..0), I1(9..8)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte2 senden
            Byte = (S1_L << 4) + (I2_H << 2) + (I2_L >> 6);           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 3: S1(3..0), I2(9..6)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte3 senden
            Byte = (S2_L << 6) + (S1_H << 4) + (S1_L >> 4);           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 4: S2(1..0), S1(9..4)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte4 senden
            Byte = (S2_H << 6) + (S2_L >> 2);           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 5: S2(9..2)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte5 senden
            Byte = V1_L;           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 6: V1(7..0)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte6 senden
            Byte = (V2_L << 2) + V1_H;           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 7: V2(5..0), V1(9..8)
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte7 senden
            Byte = (V2_H << 2) + (V2_L >> 6);           //Whrend bertragen wird, nchstes Byte vorbereiten: Byte 8: L4, L3, L2, L1, V2(9..6)
            if (LB & 0x40) {Byte = Byte + 0x10;}
            if (LC & 0x20) {Byte = Byte + 0x20;}
            if (LC & 0x10) {Byte = Byte + 0x40;}
            if (LC & 0x40) {Byte = Byte + 0x80;}
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = Byte;           //Byte8 senden
            while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
            TXREG = 0x00;           //Frame-Stop = 0x00 senden
        }
        
        //Warte bis Sample-Rate-Counter berluft
        while (wait>0) {
            while (!PIR1bits.TMR2IF);   //Warte bis TMR2 berluft
            PIR1bits.TMR2IF = 0;        //berlauf-Flag lschen und nchsten Messzyklus starten
            if (PIR1bits.RCIF) UART_Receive();  //Wenn zwischenzeitlich ein Anruf ber UART-Rx empfangen wurde, diesen abarbeiten
            wait--;
        }
    }   //Ende der while(1)-Schleife, nchsten Messzyklus starten
    return;
}

void UART_Receive() {
    //Ein Anruf ber UART wurde empfangen
    Byte = RCREG;
    if (Byte == 0x00) {          //Kommando 0x00: Neuer Wert fr Sample-Rate-Generator
        TXREG = Byte;               //Kommando zurcksenden als Aufforderung, nchstes Byte zu schicken
        PIR1bits.RCIF = 0;          //Receive-Flag lschen um fr nchstes Empfangsbyte bereit zu sein
        while (!PIR1bits.RCIF);     //Auf nchstes Byte warten (neuer PR2-Wert)
        WaitCnt = RCREG;            //WaitCnt setzen
        TXREG = WaitCnt;            //Byte zur Kontrolle zurcksenden und als Besttigung dass WaitCnt neu gesetzt wurde
    } else if (Byte == 0x01){   //Kommando 0x01: Starte Messungen und Messbertragungen
        TXREG = Byte;               //Kommando zurcksenden als Besttigung, dass Messungen gestartet werden
        PIR1bits.RCIF = 0;          //Receive-Flag lschen um fr nchstes Empfangsbyte bereit zu sein
        while (!PIR1bits.RCIF);     //Auf nchstes Byte warten (0x01 als finales Go)
		Byte = RCREG;            	//Dummy-Lesen, damit das Byte aus dem Buffer kommt (wird nicht auf Korrektheit geprft)
        Flags = Flags | 0x01;       //Run-Flag setzen
    } else if (Byte == 0x02){   //Kommando 0x02: Stoppe Messungen und Messbertragungen    
        UART_SendStopSeq();         //10x 0x02 zurcksenden als Besttigung, dass Messungen gestoppt werden
        Flags = Flags & 0xFE;       //Run-Flag lschen
    } else if (Byte == 0x03){   //Kommando 0x03: Liefere aktuellen Wert des Sample-Rate-Generators zurck
        TXREG = WaitCnt;                //PR2 zurcksenden    
    } else if (Byte == 0x04){   //Kommando 0x04: Liefere aktuelle Flags zurck
        TXREG = Flags;              //Flags zurcksenden  
    } else {                    //Unbekanntes Kommando
        TXREG = 0xFF;               //Unknown = 0xFF zurcksenden
    }
    PIR1bits.RCIF = 0;      //Receive-Flag lschen um fr nchsten Anruf bereit zu sein
}

void UART_SendStopSeq() {
    //Sendet Stop Besttigunssequenz (10x 0x02) ber BT
    //10x deswegen, weil Anzeigegert aktuell 10er-Pakete erwartet
    for (b=10;b>0;b--) {
        while(!TXSTAbits.TRMT); //Warte bis letzter Transmit beendet ist
        TXREG = 0x02;           //0x02 senden
    }
}

void GPIO_Init() {
    //Aufsetzen der GPIOs
    //TRISA0 = 0;             //RA0=Output (Pin ist unbelegt/floating)
    //TRISA1 = 0;             //RA1=Output (Pin ist unbelegt/floating)  
    RA0 = 0;
    RA1 = 0;
    TRISB7 = 0;             //RB7(UART-Tx)=Output  
    TRISB5 = 1;             //RB5(UART-Rx)=Input 
    TRISB6 = 1;             //Logik-Eingnge L1-L4, jeweils Input
    TRISC5 = 1;
    TRISC4 = 1;
    TRISC6 = 1;
    ANSELbits.ANS3 = 0;     //Kein ADC an AN3
    ANSELbits.ANS4 = 1;     //AN4 ist ADC-Eingang
    ANSELbits.ANS5 = 1;     //AN5 ist ADC-Eingang
    ANSELbits.ANS6 = 1;     //AN6 ist ADC-Eingang
    ANSELbits.ANS7 = 1;     //AN7 ist ADC-Eingang
    ANSELHbits.ANS8 = 0;    //Kein ADC an AN8
    ANSELHbits.ANS9 = 1;    //AN9 ist ADC-Eingang
    ANSELHbits.ANS10 = 1;   //AN10 ist ADC-Eingang
    ANSELHbits.ANS11 = 0;   //Kein ADC an AN11
}

void ADC_Init() {
    //Aufsetzen des ADC
    ADCON0bits.ADON = 1;    //ADC einschalten
    ADCON0bits.CHS = 7;     //Channel-Select AN4
    ADCON1bits.NVCFG0 = 0;  //VRef- = GND
    ADCON1bits.NVCFG1 = 0;
    ADCON1bits.PVCFG0 = 0;  //VRef+ = 5V
    ADCON1bits.PVCFG0 = 0;
    ADCON2bits.ADCS = 5;    //Fosc/16, d.h. 1,33us AD-Zeit 
    ADCON2bits.ACQT = 4;    //8 Tad-Zeiten fr Charge sollten fr alle Flle reichen
    ADCON2bits.ADFM = 1;    //Right justified, d.h.ADRESL beinhaltet Bits 7...0, ADRESH beinhaltet Bits 9..8
}

void TMR2_Init() {
    //Aufsetzen des Sample-Rate Generators
    PR2=49;                //PR2 initialisieren auf 250Hz
    T2CONbits.T2OUTPS = 14; //Post-Scale = 1:15
    T2CONbits.T2CKPS = 3;   //Pre-Scale = 1:16
    T2CONbits.TMR2ON = 1;   //TMR2 einschalten
}

void UART_Init() {
    //PIC18F13K50 UART Konfiguration
    TXSTAbits.TX9 = 0;      //8-Bit
    TXSTAbits.TXEN = 1;     //Enable Transmit        
    TXSTAbits.SYNC = 0;     //Async Mode
//    TXSTAbits.BRGH = 1;     //Baudrate High Speed
    TXSTAbits.BRGH = 0;     //Baudrate Low Speed
    RCSTAbits.SPEN = 1;     //Enable Serial Port
    RCSTAbits.RX9 = 0;      //8-Bit
    RCSTAbits.CREN = 1;     //Enable reception bit Rx
    BAUDCONbits.DTRXP = 0;  //Keine Invertierung (BT-Modul liefert Active-High)
    BAUDCONbits.CKTXP = 0;  //Keine Invertierung (BT-Modul erwartet Active-High) 
    BAUDCONbits.BRG16 = 1;  //16-Bit Baudrate-Gernerierung
    BAUDCONbits.WUE = 0;    //Normaler Receive ohne Wake-up
    BAUDCONbits.ABDEN = 0;  //Ohne Auto-Baud-Detect
//    SPBRGH = 0;             //16-Bit Baudzhler bei BRGH=1: 
//    SPBRG = 25;             //BRG = (12MHz/(115200*4))-1 = 25
//    SPBRGH = 0;             //16-Bit Baudzhler bei BRGH=0: 
//    SPBRG = 12;             //BRG = (12MHz/(57600*16))-1 = 12
//    SPBRGH = 0;             //16-Bit Baudzhler bei BRGH=0: 
//    SPBRG = 38;             //BRG = (12MHz/(19200*16))-1 = 38
    SPBRGH = 0;             //16-Bit Baudzhler bei BRGH=0: 
    SPBRG = 19;             //BRG = (12MHz/(38400*16))-1 = 19
}

